//============================================================================ 
// 
//		 Program/Module "Parametric Linear Solver" 
// 
// 
//	  Copyright (c) 2003-2010   Evgenija Popova, Michael Zimmer
// 
// This program/module is free software for non-commercial use. 
// For details on theory  see the papers 
// 
// S. Rump: Verification methods for dense and sparse systems of equations, 
// in: Topics in Validated Computations, ed. J. Herzberger, Elsevier Science 
// B.V., 1994,	pp. 63-135. 
//
// E. Popova, W. Kraemer: Parametric Fixed-point Iteration Implemented in 
// C-XSC,  Universitaet Wuppertal, Preprint BUW - WRSWT, 2003.
// 
// This program/module is distributed WITHOUT ANY WARRANTY. 
// 
//============================================================================ 
//---------------------------------------------------------------------------- 
// File: parlinsys (implementation) 
// 
// Purpose: Computing a verified inclusion for the parametric solution set 
// of a square parametric linear system of equations A(p)*x = b(p) with full 
// parametric matrix A(p) and parametric right-hand side b(p) which elements 
//             are affine-linear combinations of k given interval parameters. 
// Method: Transformation of A(p)*x = b(p) to fixed-point form and applying 
//  	    an interval residual iteration. 
// 
// Global functions: 
//    ParLinSolve()	 : to get a verified enclosure of the solution 
//    ParLinSolveErrMsg(): to get an error message text 
// 
//---------------------------------------------------------------------------- 

#include <string.h>         
#include <imatrix.hpp>      // Interval matrix/vector arithmetic 
#include <mvi_util.hpp>     // Interval matrix/vector utility functions 
#include <matinv_aprx.hpp>  // Matrix inversion 
#include "parlinsys.hpp"    
#include "timer.hpp"

#ifdef _OPENMP
#include <omp.h>
#endif
 
using namespace cxsc; 
using namespace std; 
 
//---------------------------------------------------------------------------- 
// Error codes used in this module. 
//---------------------------------------------------------------------------- 
 
const int 
  NoError      = 0,   // No error occurred 
  NotSquare    = 1,   // System to be solved is not square 
  DimensionErr = 2,   // Dimensions of A and b are not compatible 
  InvFailed    = 3,   // System is probably singular 
  VerivFailed  = 4;   // Verification failed, system is probably 
                      // ill-conditioned 
 
//---------------------------------------------------------------------------- 
// Error messages depending on the error code. 
//---------------------------------------------------------------------------- 
 
char* ParLinSolveErrMsg (int Err) 
{ 
  static char Msg[80] = ""; 
 
  if (Err != NoError) { 
    char Hlp[60]; 
 
    switch (Err) { 
      case NotSquare: 
        strcpy(Hlp,"System to be solved is not square"); break; 
      case DimensionErr: 
        strcpy(Hlp,"Dimensions of A and b are not compatible"); break; 
      case InvFailed: 
        strcpy(Hlp,"System is probably singular"); break; 
     case VerivFailed: 
        strcpy(Hlp,"Verification failed, system is probably ill-conditioned"); 
        break; 
      default: 
        strcpy(Hlp,"Code not defined"); 
    } 
    sprintf(Msg,"Error: %s!",Hlp); 
  } 
  return(Msg); 
} // end  ParLinSolveErrMsg 
 
 
//! Tries to guess zero values in successive iterates
/* The vectors x and y are successive approximations for the solution of a
   linear system of equations computed by iterative refinement. If a compo-
   nent of y is diminished by more than 'Factor', it is a good candidate for
   a zero entry. Thus, it is set to zero.
   \param x An iterate
   \param y Iterate succeding x
*/
template<typename Tvec>
inline void CheckForZeros ( Tvec& x, Tvec& y )
{
  const real Factor = 1E+5;

  for (int i = Lb(y); i <= Ub(y); i++)
    if ( abs(y[i])*Factor < abs(x[i]) ) y[i] = 0.0;
}


//! Checks if an iterate is sufficiently accurate
/*
   The vectors x and y are successive iterates. The function returns true if
   the relative error of all components x_i and y_i is <= 10^(-12), i.e. y_i
   has about 12 correct decimals. If x_i or y_i vanishes, the relative error
   is implicitly set to zero.
   \param x An iterate
   \param y Iterate succeding x
   \return true, if y hast about 12 correct decimals, else false
*/
template<typename Tvec>
inline int Accurate ( const Tvec& x, const Tvec& y )
{
  const real Delta = 1E-12;   // Relative error bound
  int        i, n, ok;

  i = Lb(y); n = Ub(y);
  do {
    // Relative error > Delta ?
    ok = (abs(y[i] - x[i]) <= Delta * abs(y[i]) );
    i++;
  } while (ok && (i <= n));

  return ok;
} // Accurate


//C-XSC Blow-Function for class civector does not exist, this function is a
//workaround
inline civector Blow(const civector &x, const real &eps) {
  civector y(Lb(x), Ub(x));
  
  for(int i=Lb(x) ; i<=Ub(x) ; i++)
    y[i] = Blow(x[i], eps);
  
  return y;
}

//---------------------------------------------------------------------------- 
// This function 'VerificationStep()' performs an iteration with  
// Einzelschrittverfahren: for k = 1,2,...  
// [y] = [x] = Blow([x],Epsilon), [x]_i = [z]_i + [C]_i*[x], i = 1,..n until 
// the new iterate [x] lies in [y] or until the maximum number of iterations  
// is exceeded. The factor 'Epsilon' for the inflation is specified by the
// user. The flag 'IsVerified' is set if an inclusion could be established. 
//---------------------------------------------------------------------------- 
template<class Tvec, class Tmat>
static void VerificationStep(Tvec& xx, Tvec& zz, 
                              Tmat& C, int& IsVerified) 
{ 
  const int  MaxIter = 10;          // Maximal number of iteration steps 
  const real Epsilon = 0.1;         // Epsilon inflation 
 
  int     p, i; 
  Tvec yy(Lb(xx), Ub(xx)); 
 
  xx = zz; p = 0;                   // Initialize:  [x] = [z] 
  do  
  {	   
    yy = xx = Blow(xx, Epsilon);          // Epsilon inflation 
    for(i = Lb(xx); i <= Ub(xx); i++)     // Einzelschrittverfahren  
      xx[i] = zz[i] + C[i] * xx;          // new iterate  
    IsVerified = in(xx, yy);              // Inclusion in the interior??? 
    p++; 
  }  
  while (!IsVerified && (p < MaxIter)); 

}

//----------------------------------------------------------------------------

//'maxDist()' computes the maximal distance between two 
// real/complex interval vectors.
// For interval x, y; dist(x,y)= Max(Abs(x.inf-y.inf), Abs(x.sup-y.sup))
//----------------------------------------------------------------------------
template<class TVec>
static real maxDist(const TVec& xx, const TVec& yy) {
  real dist = 0.0, tmp;
  int n = VecLen(xx);

  for(int i=1 ; i<=n ; i++) {
    tmp = abs(Inf(xx[i]) - Inf(yy[i]));
    if(tmp > dist) dist = tmp;
    tmp = abs(Sup(xx[i]) - Sup(yy[i]));
    if(tmp > dist) dist = tmp;
  }

  return dist;
}

//----------------------------------------------------------------------------

//'RefinementStep()' performs iterative refinement of a solution enclosure [x]: 
// p= 1,2,... [y]= [x], [x]= [z] + [C].[x], [x] &= [y], until 
// the distance between the two iterates [x] and [y] is > Epsilon 
// or until the maximum number of iterations is exceeded. 
// The factor 'Epsilon' for the distance between two iterates is user specified. 
// [z] and [C] are the same as in the Verification step.
// 
//----------------------------------------------------------------------------

template<class TVec, class TMat> 
static void RefinementStep(TVec& xx, const TVec& zz, const TMat& C, const real& Epsilon) { 
  const int  MaxIter = 20;          // Maximal number of iteration steps 
  int        p=0; 
  TVec yy(Lb(xx), Ub(xx)); 
 
  do {	   
    //cout << "Refinement Step " << p+1 << endl;
    yy = xx;
    xx = zz + C*xx;
    xx &= yy;
    p++;
  } while (maxDist(xx, yy) > Epsilon && (p < MaxIter)); 

}


// This function 'Dual' reverses the end-points of an interval.
// ------------------------------------------------------------
static void Dual(interval& xx){ 
  real tmp;
    
  tmp = Inf(xx);
  UncheckedSetInf(xx, Sup(xx));
  UncheckedSetSup(xx, tmp);
} // End Dual


//Computes an inner enclosure of the true result of the linear system
static void InnerEstimation(int n, int p, ivector& yy, const ivector& ip, const imatrix& T, const imatrix& C, 
                            const ivector& zz, const rvector& x, const ivector& xx) {
  ivector cp(p);
  Resize(yy,n);
  dotprecision Accu;
  idotprecision IAccu;

  for (int j = 1; j <= n; j++) {

    cp = ip;

    for (int i=1; i<=p; i++)  {

      if ( Inf(T[j][i]) >= 0 )  { 

        // T>=0
        Dual(cp[i]);
        if ( Sup(ip[i]) <= 0 ) Dual(T[j][i]);  // ip <= 0
        else if ( Inf(ip[i]) < 0 ) UncheckedSetSup(T[j][i], Inf(T[j][i]));

      } else if ( Sup(T[j][i]) <= 0 )  {

        // T<=0
        if ( Sup(ip[i]) <= 0 ) Dual(T[j][i]);
        else if ( Inf(ip[i]) < 0 ) UncheckedSetInf(T[j][i], Sup(T[j][i]));

      } else { 

        // 0 in T 
        if ( Inf(ip[i]) < 0 && Sup(ip[i]) > 0 ) cp[i] = interval(0);
        else if ( Sup(ip[i]) <= 0 ) {  
          Dual(T[j][i]);
	  UncheckedSetInf(cp[i], Sup(ip[i]));
        } else {
          UncheckedSetSup(cp[i], Inf(ip[i]));
        }
      }
   }

   Accu = 0.0;
   accumulate(Accu, Inf(T[j]), Inf(cp));
   UncheckedSetInf(zz[j], rnd(Accu, RND_DOWN));   

   Accu = 0.0;
   accumulate(Accu, Sup(T[j]), Sup(cp));
   UncheckedSetSup(zz[j],  rnd(Accu, RND_UP));

   IAccu = x[j];
   IAccu += zz[j]; 
   accumulate(IAccu, C[j], xx);
   yy[j] = rnd(IAccu);
   Dual(yy[j]);
 }  // end for_j

}
 
//---------------------------------------------------------------------------- 
// Purpose: The function 'ParLinSolve()' computes a verified enclosure for 
// the parametric solution set of a square parametric linear system of  
// equations A(p)*x=b(p), having a factorized by the parameters representation: 
//    (A0+[p1]A1+[p2]A2+ +[pk]Ak)x = (b0+[p1]b1+[p2]b2+ +[pk]bk).
// Parameters: 
//    In : 'Ap'         : coefficient matrix of the system in a factorized  
//                        representation Ap = (A0,A1,..,Ak) from R(n(k+1),n); 
//                        the matrix is passed as reference parameter to  
//                        save time for copying it; 
//         'bp'         : coefficients right-hand side of the system in a  
//                        factorized matrix representation 
//                        bp = (b0,b1,..bk) from R(n,k+1); matrix bp is passed  
//                        as reference parameter to save time for copying it; 
//         'ip'         : vector of the interval values for the parameters, 
//                        passed as reference parameter to save time for 
//                        copying it; 
//         'SharpC'     : flag signalling whether a sharp enclosure of the  
//                        contracting matrix C is to be computed; 
//         'Inner'      : flag signalling whether inner estimation of the hull
//			  of the solution set is to be computed;
//         'prec'       : precision to use for dot product computations
//                        (note: for performance reasons, some computations are
//                        always executed with precision 1 (double precision))
//    Out: 'xx'         : enclosure of the parametric solution set, resized to 
//                        standard index range with lower index bound 1; 
//	   'yy'		: inner estimation of the exact hull for the parametric
//			  solution set, resized to standard index range with 
//			  lower index bound 1;
//         'Err'        : error code; 
// Description: An approximate inverse 'R' of the mid-point matrix is computed  
//   by calling function 'MatInv()'. Then an approximate mid-point solution 'x' 
//   is computed applying a conventional real residual iteration. For the final 
//   verification, an interval residual iteration is performed. An enclosure of 
//   the unique solution is returned in the interval vector 'xx' and an inner  
//   approximation of the hull of the solution set is returned in the interval
//   vector 'yy'.
//---------------------------------------------------------------------------- 
 void ParLinSolveMain( rmatrix&  Ap, rmatrix&  bp, ivector&  ip,  
                       ivector&  xx, bool  SharpC, bool Inner,
		       ivector&  yy, int&  Err, int prec ) 
{ 
  const int MaxResCorr = 10;    // Maximal number of real residual corrections 
  const real Epsilon   = 1e-10;
 
  int     p = VecLen(ip);            // Length of the parametric vector 'ip' 
				     // here it is of length p=p+1 for p parameters
  int     n = RowLen(Ap);            // Length of the rows of 'Ap' 
  int     m = ColLen(Ap);            // Length of the columns of 'Ap' 
  int     IsVerified, i, j, k; 
  int     ARowLow, AColLow,          // Lower index bounds of 'Ap', 'bp' 
          bRowLow, bColLow, ipLow;   // and ip 
  int     oldprec = opdotprec;
   
  rmatrix R(n,n);                         // To store the inverse of 'mid(A([ip]))' 
  double start,end;
   
  start = GetTime();

  if (m != n*p)                      // Error: 'A(p)' is not square 
    { Err = NotSquare; return; } 
 
  if (n != ColLen(bp) &&  
         p != RowLen(bp))            // Error: Dimensions of 'A(p)' and  
    { Err = DimensionErr; return; }  //       'b(p)' are not compatible 
 
 
// Start algorithm 
//---------------- 
  rvector       x(n), y(n), d(n);  // Allocate dynamic vectors 
  imatrix       C(n, n);           // Allocate dynamic interval matrix C
  ivector       zz(n);	           // Allocate dynamic nonparametric vector 
  dotprecision  Accu;              // Allocate accumulators 
  idotprecision IAccu; 
 
  Accu.set_dotprec(prec);
  IAccu.set_dotprec(prec);
  opdotprec = prec;

  C = 0.0;
  
// Normalize index range of 'Ap' and 'ip' to standard range 1..n 
// Save lower bounds of 'ip' and 'Ap' 
// to restore them before leaving 'ParLinSolve()'. 
//----------------------------------------------------------------- 
  ARowLow = Lb(Ap, ROW); SetLb(Ap, ROW, 1); 
  AColLow = Lb(Ap, COL); SetLb(Ap, COL, 1); 
  ipLow = Lb(ip); SetLb(ip, 1); 


 rmatrix    Am(n, n);               // Allocate dynamic mid-point matrix
 rvector    bm(n), pm(p);           // Allocate dynamic mid-point vectors
 
// Compute mid-point vectors/matrix: pm, Am, bm 
//---------------------------------------------------------------------- 
  pm = mid(ip); 
  bm = bp * pm;
  Am = Ap(1, n, 1, n); 
  for (i = 2; i <= p; i++)      
    Am += pm[i]* Ap(n*(i-1)+1, i*n, 1, n); 
//---------------------------------------------------------------------- 

  end = GetTime();
  //cout << "Preparation: " << end-start << endl;
 
  start = GetTime();
  MatInvAprx(Am, R, Err); 
  end = GetTime();
  //cout << "Computation of approximate inverse: " << end-start << endl;
  if (Err != NoError)                   // Error: Inversion failed 
    { Err = InvFailed; return; } 

  
// Normalize index range of 'R'  to standard index range 1..n 
//---------------------------------------------------------------------
  start = GetTime();
  SetLb(R, ROW, 1); SetLb(R, COL, 1);
 
  x = R * bm; k = 0;   // Real residual iteration 
  do  
  {                
    y = x; 
    for (i = 1; i <= n; i++)  
    {                               // Compute: d = #*(bm-Am*y) 
      Accu = bm[i];                 //----------------------- 
      accumulate(Accu, -Am[i], y); 
      d[i] = rnd(Accu); 
    } 
    
    for (i = 1; i <= n; i++)        // Compute: x = #*(y+R*d) 
    {                               //----------------------  
      Accu = y[i];                
      accumulate(Accu, R[i], d); 
      x[i] = rnd(Accu); 
    } 
     
    CheckForZeros(y, x); 
    k++; 
  }  
  while (!Accurate(y, x) && (k < MaxResCorr)); 

 
  end = GetTime();
  //cout << "Defect iteration: " << end-start << endl;

//  Resize mid-point vectors and matrix to save memory
//  ----------------------------------------------------------------
    Resize(pm);  Resize(bm);  Resize(Am);
    Resize(y);

// Prepare verification step: 

  start = GetTime();
  imatrix T;	// Interval matrix for intermediate quantities
 
// compute enclosure [C]  of   I - R*A(p) for p in [ip]  
//--------------------------------------------------------------
 
  Accu.set_dotprec(1);
  IAccu.set_dotprec(1);
  opdotprec = 1;
  if (SharpC) {
      // Sharp enclosure C = I - R*A(p) 

      #ifdef CXSC_USE_BLAS

        for(i=1 ; i<=n ; i++)
           C[i][i] = 1.0;

        #ifdef _OPENMP
          #pragma omp parallel private(j,k) default(shared)
          {

            int size = n/omp_get_num_threads();
            #pragma omp for 
            for(j=1 ; j<=omp_get_num_threads() ; j++) {
              for(k=1 ; k<=p ; k++) {
                C((j-1)*size+1,j*size,1,n) -= ( R((j-1)*size+1, j*size, 1, n) * imatrix(Ap(n*(k-1)+1, n*k, 1, n)) ) * ip[k];
              } 
            }

            //Compute rest if matrix not evenly distributable among threads
            #pragma omp single
            {
              if(n % omp_get_num_threads() != 0) {
                for(k=1 ; k<=p ; k++) {
                  C(size*omp_get_num_threads()+1,n,1,n) -= ( R(size*omp_get_num_threads()+1, n, 1, n) * imatrix(Ap(n*(k-1)+1, n*k, 1, n)) ) * ip[k];
                }
              }
            }

          }

        #else 

          for(k=1 ; k<=p ; k++) 
             C -= (R * imatrix(Ap(n*(k-1)+1, n*k, 1, n))) * ip[k];

        #endif //ifdef _OPENMP

      #else

        Resize(T, 1, p);       // Allocate dynamic temporary matrix T
  
        #ifdef _OPENMP
        #pragma omp parallel for private(i,j,k) firstprivate(Accu,IAccu,T) default(shared)
        #endif    
        for (i = 1; i <= n; i++) 
         for (j = 1; j <= n; j++) {                                     
           Accu = (i == j) ? 1.0 : 0.0;         
           accumulate(Accu, -R[i], Ap(1, n, 1, n)[Col(j)]); 
           rnd(Accu, T[1][1]); 
           for (k = 2;k <= p; k++) { 
             Accu = 0.0; 
	     accumulate(Accu, -R[i], Ap(n*(k-1)+1, n*k, 1, n)[Col(j)]); 
	     rnd(Accu, T[1][k]); 
           } 
           IAccu = 0.0; 
           accumulate(IAccu, T[Row(1)], ip); 
           C[i][j] = rnd(IAccu); 
         } 

      #endif

  } else {

      // Rough enclosure C = I - R*A([ip])  

      #ifdef CXSC_USE_BLAS

        Resize(T, n, n);     // Allocate dynamic temporary matrix T 
        T = Ap(1, n, 1, n);  //                  to contain A([ip])
  
        #ifdef _OPENMP
    
        #pragma omp parallel for private(i) default(shared)
        for(int r=1 ; r<=n ; r++) {
          for(int c=1 ; c<=n ; c++)
            for (i=2; i<=p; i++) 
                T[r][c] += ip[i]* Ap[r+n*(i-1)][c]; 
        }

        #else 

          for (i = 2; i <= p; i++)
          T += ip[i]* Ap(n*(i-1)+1, i*n, 1, n); 

        #endif //ifdef _OPENMP

        for(i=1 ; i<=n ; i++)
          C[i][i] = 1.0;

        C -= R*T;      

      #else

        Resize(T, n, n);     // Allocate dynamic temporary matrix T 
        T = Ap(1, n, 1, n);  //                  to contain A([ip])
        for (i = 2; i <= p; i++)
           T += ip[i]* Ap(n*(i-1)+1, i*n, 1, n); 

        #ifdef _OPENMP
        #pragma omp parallel for private(i,j) firstprivate(IAccu) default(shared)
        #endif
        for (i = 1; i <= n; i++) 
         for (j = 1; j <= n; j++) {                                       
           IAccu = (i == j) ? 1.0 : 0.0;         
           accumulate(IAccu, -R[i], T[Col(j)]); 
           rnd(IAccu, C[i][j]); 
         } 

      #endif //ifdef CXSC_USE_BLAS

  }
  Accu.set_dotprec(prec);
  IAccu.set_dotprec(prec);
  opdotprec = prec;

   end = GetTime();
   //cout << "Computation of [C]: " << end-start << endl;

// Resize T to length (n,p).
// Normalize index range of 'bp' to standard range 1..n.
// Save lower bounds of 'bp'
// to restore them before leaving 'ParLinSolve()'.
//---------------------------------------------------------------------
   Resize(T, n, p);
   bRowLow = Lb(bp, ROW); SetLb(bp, ROW, 1);
   bColLow = Lb(bp, COL); SetLb(bp, COL, 1);   
 
// compute enclosure [z] of z = R*(b(p)-A(p)*x). 
//---------------------------------------------------------------------
  start = GetTime();
  #ifdef _OPENMP
  #pragma omp parallel for private(k,i,j,m) firstprivate(Accu,IAccu,d,zz) default(shared)
  #endif
  for (k = 1; k <= p; k++) { 
 
   j = n*(k-1); 

    for (i = 1; i <= n; i++)               // Compute: d = #*(b_k - A_k * x) 
    {                                      //------------------------------ 
      Accu = bp[Col(k)][i]; 
      m = j + i;                 
      accumulate(Accu, -Ap[m], x);  
      d[i] = rnd(Accu); 
    } 
 
    for (i = 1; i <= n; i++)                // Compute: [y] = ##(b_k - A_k*x - d) 
    {                                       //----------------------------------  
      Accu = bp[Col(k)][i];      
      m = j + i;                             
      accumulate(Accu, -Ap[m], x); 
      Accu -= d[i]; 
      rnd(Accu, zz[i]); 
    } 
 
                                //  Now b_k -A_k*x lies in d + [y].  
                                //  Thus R*(b_k-A_k*x) lies in [z]= ##(R*d + R*[y])     
    for (i = 1; i <= n; i++)  
    {   
      IAccu = 0.0; 
      accumulate(IAccu, R[i], d); 
      accumulate(IAccu, R[i], zz); 
      T[Col(k)][i] = rnd(IAccu); 
    } 
 
  }  // end for_k 
 
  #ifdef _OPENMP
  #pragma omp parallel for private(i) firstprivate(IAccu) default(shared)
  #endif
  for(i = 1; i <= n; i++) 
  { 
    IAccu = 0.0; 
    accumulate(IAccu, T[i], ip);   
    zz[i] = rnd(IAccu); 
  } 
   
  end = GetTime();
  //cout << "Computation of [z]: " << end-start << endl;

 
// Resize 'xx' to length 'n'.
// ---------------------------------- 
   Resize(xx, n);  
  

// verified enclosure [x] for the absolute error of x. 
//-------------------------------------------------------------------------
   
//  VerificationStep(xx, zz, C, IsVerified);   // Attempt to compute [x] 
  start = GetTime();
  VerificationStep(xx, zz, C, IsVerified); 
  if (!IsVerified) 
      Err = VerivFailed; 
  else 
  {
    if (Inner)        // Compute Inner Estimation of the outer solution
    {		      // --------------------------------------------------
      start = GetTime();
      InnerEstimation(n, p, yy, ip, T, C, zz, x, xx);
      end = GetTime();
      //cout << "Inner Estimation: " << end-start << endl;
    }    // end Inner Estimation of the solution enclosure. 

    //RefinementStep(xx,zz,C,Epsilon);
    xx = x + xx;             // The solution set lies in x+[x]
   }

   end = GetTime();
   //cout << "Verification: " << end-start << endl; 

  // Restore lower index bounds of 'Ap', 'bp' and ip 
  //-------------------------------------------------   
  SetLb(bp, ROW, bRowLow); SetLb(bp, COL, bColLow); 
  SetLb(Ap, ROW, ARowLow); SetLb(Ap, COL, AColLow); 
  SetLb(ip, ipLow); 

  opdotprec = oldprec;
 
} //ParLinSolve 


//---------------------------------------------------------------------------- 
// Purpose: The function 'ParLinSolve()' computes a verified enclosure for 
// the parametric solution set of a square parametric linear system of  
// equations A(p)*x=b(p), having a factorized by the parameters representation: 
//    (A0+[p1]A1+[p2]A2+ +[pk]Ak)x = (b0+[p1]b1+[p2]b2+ +[pk]bk).
// Parameters: 
//    In : 'Ap'         : coefficient matrix of the system in a factorized  
//                        representation Ap = (A0,A1,..,Ak) from R(n(k+1),n); 
//                        the matrix is passed as reference parameter to  
//                        save time for copying it; 
//         'bp'         : coefficients right-hand side of the system in a  
//                        factorized matrix representation 
//                        bp = (b0,b1,..bk) from R(n,k+1); matrix bp is passed  
//                        as reference parameter to save time for copying it; 
//         'ip'         : vector of the interval values for the parameters, 
//                        passed as reference parameter to save time for 
//                        copying it; 
//         'SharpC'     : flag signalling whether a sharp enclosure of the  
//                        contracting matrix C is to be computed; 
//         'Inner'      : flag signalling whether inner estimation of the hull
//			  of the solution set is to be computed;
//         'prec'       : precision to use for dot product computations
//                        (note: for performance reasons, some computations are
//                        always executed with precision 1 (double precision)
//    Out: 'xx'         : enclosure of the parametric solution set, resized to 
//                        standard index range with lower index bound 1; 
//	   'yy'		: inner estimation of the exact hull for the parametric
//			  solution set, resized to standard index range with 
//			  lower index bound 1;
//         'Err'        : error code; 
// Description: An approximate inverse 'R' of the mid-point matrix is computed  
//   by calling function 'MatInv()'. Then an approximate mid-point solution 'x' 
//   is computed applying a conventional real residual iteration. For the final 
//   verification, an interval residual iteration is performed. An enclosure of 
//   the unique solution is returned in the interval vector 'xx' and an inner  
//   approximation of the hull of the solution set is returned in the interval
//   vector 'yy'.
//---------------------------------------------------------------------------- 
 void ParLinSolveMain( cmatrix&  Ap, cmatrix&  bp, civector&  ip, civector&  xx, 
                       bool  SharpC, int&  Err, int prec ) 
{ 
  const int MaxResCorr = 10;    // Maximal number of real residual corrections 
  const real Epsilon   = 1e-10;
 
  int     p = VecLen(ip);            // Length of the parametric vector 'ip' 
				     // here it is of length p=p+1 for p parameters
  int     n = RowLen(Ap);            // Length of the rows of 'Ap' 
  int     m = ColLen(Ap);            // Length of the columns of 'Ap' 
  int     IsVerified, i, j, k; 
  int     ARowLow, AColLow,          // Lower index bounds of 'Ap', 'bp' 
          bRowLow, bColLow, ipLow;   // and ip 
  int     oldprec = opdotprec;
   
  cmatrix R(n,n);                         // To store the inverse of 'mid(A([ip]))' 
  double start,end;
   
  start = GetTime();

  if (m != n*p)                      // Error: 'A(p)' is not square 
    { Err = NotSquare; return; } 
 
  if (n != ColLen(bp) &&  
         p != RowLen(bp))            // Error: Dimensions of 'A(p)' and  
    { Err = DimensionErr; return; }  //       'b(p)' are not compatible 
 
 
// Start algorithm 
//---------------- 
  cvector       x(n), y(n), d(n);  // Allocate dynamic vectors 
  cimatrix       C(n, n);           // Allocate dynamic interval matrix C
  civector       zz(n);	           // Allocate dynamic nonparametric vector 
  cdotprecision  Accu;              // Allocate accumulators 
  cidotprecision IAccu; 
 
  Accu.set_dotprec(prec);
  IAccu.set_dotprec(prec);
  opdotprec = prec;

  C = 0.0;
  
// Normalize index range of 'Ap' and 'ip' to standard range 1..n 
// Save lower bounds of 'ip' and 'Ap' 
// to restore them before leaving 'ParLinSolve()'. 
//----------------------------------------------------------------- 
  ARowLow = Lb(Ap, ROW); SetLb(Ap, ROW, 1); 
  AColLow = Lb(Ap, COL); SetLb(Ap, COL, 1); 
  ipLow = Lb(ip); SetLb(ip, 1); 


 cmatrix    Am(n, n);               // Allocate dynamic mid-point matrix
 cvector    bm(n), pm(p);           // Allocate dynamic mid-point vectors
 
// Compute mid-point vectors/matrix: pm, Am, bm 
//---------------------------------------------------------------------- 
  pm = mid(ip); 
  bm = bp * pm;
  Am = Ap(1, n, 1, n); 
  for (i = 2; i <= p; i++)      
    Am += pm[i]* Ap(n*(i-1)+1, i*n, 1, n); 

//cout << Am << endl;
//---------------------------------------------------------------------- 

  end = GetTime();
  //cout << "Preparation: " << end-start << endl;
 
  start = GetTime();
  MatInvAprx(Am, R, Err); 
  end = GetTime();
  //cout << "Computation of approximate inverse: " << end-start << endl;
  if (Err != NoError)                   // Error: Inversion failed 
    { Err = InvFailed; return; } 

// Normalize index range of 'R'  to standard index range 1..n 
//---------------------------------------------------------------------
  start = GetTime();
  SetLb(R, ROW, 1); SetLb(R, COL, 1);
 
  x = R * bm; k = 0;   // Real residual iteration 
  do  
  {                
    y = x; 
    for (i = 1; i <= n; i++)  
    {                               // Compute: d = #*(bm-Am*y) 
      Accu = bm[i];                 //----------------------- 
      accumulate(Accu, -Am[i], y); 
      d[i] = rnd(Accu); 
    } 
    
    for (i = 1; i <= n; i++)        // Compute: x = #*(y+R*d) 
    {                               //----------------------  
      Accu = y[i];                
      accumulate(Accu, R[i], d); 
      x[i] = rnd(Accu); 
    } 
     
    CheckForZeros(y, x); 
    k++; 
  }  
  while (!Accurate(y, x) && (k < MaxResCorr)); 
 
  end = GetTime();
  //cout << "Defect iteration: " << end-start << endl;

//  Resize mid-point vectors and matrix to save memory
//  ----------------------------------------------------------------
    Resize(pm);  Resize(bm);  Resize(Am);
    Resize(y);

// Prepare verification step: 

start = GetTime();
cimatrix T;	// Interval matrix for intermediate quantities
 
// compute enclosure [C]  of   I - R*A(p) for p in [ip]  
//--------------------------------------------------------------
 
  Accu.set_dotprec(1);
  IAccu.set_dotprec(1);
  opdotprec = 1;
  if (SharpC) {
      // Sharp enclosure C = I - R*A(p) 

      #ifdef CXSC_USE_BLAS

        for(i=1 ; i<=n ; i++)
           C[i][i] = 1.0;

        #ifdef _OPENMP
          #pragma omp parallel private(j,k) default(shared)
          {
            int size = n/omp_get_num_threads();

            #pragma omp for 
            for(j=1 ; j<=omp_get_num_threads() ; j++) {
              for(k=1 ; k<=p ; k++) {
                C((j-1)*size+1,j*size,1,n) -= ( R((j-1)*size+1, j*size, 1, n) * cimatrix(Ap(n*(k-1)+1, n*k, 1, n)) ) * ip[k];
              } 
            }
	
            //Compute rest if matrix not evenly distributable among threads
            #pragma omp single
            {
              if(n % omp_get_num_threads() != 0) {
                for(k=1 ; k<=p ; k++) {
                  C(size*omp_get_num_threads()+1,n,1,n) -= ( R(size*omp_get_num_threads()+1, n, 1, n) * cimatrix(Ap(n*(k-1)+1, n*k, 1, n)) ) * ip[k];
                }
              }
            }

          }

        #else 

          for(k=1 ; k<=p ; k++) 
             C -= (R * cimatrix(Ap(n*(k-1)+1, n*k, 1, n))) * ip[k];

        #endif //ifdef _OPENMP

 
      #else

        Resize(T, 1, p);       // Allocate dynamic temporary matrix T
  
        #ifdef _OPENMP
        #pragma omp parallel for private(i,j,k) firstprivate(Accu,IAccu,T) default(shared)
        #endif    
        for (i = 1; i <= n; i++) 
         for (j = 1; j <= n; j++) {                                     
           Accu = (i == j) ? 1.0 : 0.0;         
           accumulate(Accu, -R[i], Ap(1, n, 1, n)[Col(j)]); 
           T[1][1] = cinterval(rnd(Accu,RND_DOWN),rnd(Accu,RND_UP)); 
           for (k = 2;k <= p; k++) { 
             Accu = 0.0; 
	     accumulate(Accu, -R[i], Ap(n*(k-1)+1, n*k, 1, n)[Col(j)]); 
	     T[1][k] = cinterval(rnd(Accu,RND_DOWN),rnd(Accu,RND_UP)); 
           } 
           IAccu = 0.0; 
           accumulate(IAccu, T[Row(1)], ip); 
           C[i][j] = rnd(IAccu); 
         } 

      #endif

  } else {

      // Rough enclosure C = I - R*A([ip])  

      #ifdef CXSC_USE_BLAS

        Resize(T, n, n);     // Allocate dynamic temporary matrix T 
        T = Ap(1, n, 1, n);  //                  to contain A([ip])
  
        #ifdef _OPENMP
    
        #pragma omp parallel for private(i) default(shared)
        for(int r=1 ; r<=n ; r++) {
          for(int c=1 ; c<=n ; c++)
            for (i=2; i<=p; i++) 
                T[r][c] += ip[i]* Ap[r+n*(i-1)][c]; 
        }

        #else 

          for (i = 2; i <= p; i++)
          T += ip[i]* Ap(n*(i-1)+1, i*n, 1, n); 

        #endif //ifdef _OPENMP

        for(i=1 ; i<=n ; i++)
          C[i][i] = 1.0;

        C -= R*T;      

      #else

        Resize(T, n, n);     // Allocate dynamic temporary matrix T 
        T = Ap(1, n, 1, n);  //                  to contain A([ip])
        for (i = 2; i <= p; i++)
           T += ip[i]* Ap(n*(i-1)+1, i*n, 1, n); 

        #ifdef _OPENMP
        #pragma omp parallel for private(i,j) firstprivate(IAccu) default(shared)
        #endif
        for (i = 1; i <= n; i++) 
         for (j = 1; j <= n; j++) {                                       
           IAccu = (i == j) ? 1.0 : 0.0;         
           accumulate(IAccu, -R[i], T[Col(j)]); 
           rnd(IAccu, C[i][j]); 
         } 

      #endif //ifdef CXSC_USE_BLAS

  }
  Accu.set_dotprec(prec);
  IAccu.set_dotprec(prec);
  opdotprec = prec;

   end = GetTime();
   //cout << "Computation of [C]: " << end-start << endl;

// Resize T to length (n,p).
// Normalize index range of 'bp' to standard range 1..n.
// Save lower bounds of 'bp'
// to restore them before leaving 'ParLinSolve()'.
//---------------------------------------------------------------------
   Resize(T, n, p);
   bRowLow = Lb(bp, ROW); SetLb(bp, ROW, 1);
   bColLow = Lb(bp, COL); SetLb(bp, COL, 1);   
 
// compute enclosure [z] of z = R*(b(p)-A(p)*x). 
//---------------------------------------------------------------------
  start = GetTime();

  #ifdef _OPENMP
  #pragma omp parallel for private(k,i,j,m) firstprivate(Accu,IAccu,d,zz) default(shared)
  #endif
  for (k = 1; k <= p; k++) { 
 
   j = n*(k-1); 

    for (i = 1; i <= n; i++)               // Compute: d = #*(b_k - A_k * x) 
    {                                      //------------------------------ 
      Accu = bp[Col(k)][i]; 
      m = j + i;                 
      accumulate(Accu, -Ap[m], x);  
      d[i] = rnd(Accu); 
    } 
 
    for (i = 1; i <= n; i++)                // Compute: [y] = ##(b_k - A_k*x - d) 
    {                                       //----------------------------------  
      Accu = bp[Col(k)][i];      
      m = j + i;                             
      accumulate(Accu, -Ap[m], x); 
      Accu -= d[i]; 
      SetRe(zz[i], rnd(Re(Accu))); 
      SetIm(zz[i], rnd(Im(Accu))); 
    } 
 
                                //  Now b_k -A_k*x lies in d + [y].  
                                //  Thus R*(b_k-A_k*x) lies in [z]= ##(R*d + R*[y])     
    for (i = 1; i <= n; i++)  
    {   
      IAccu = 0.0; 
      accumulate(IAccu, R[i], d); 
      accumulate(IAccu, R[i], zz); 
      T[Col(k)][i] = rnd(IAccu); 
    } 
 
  }  // end for_k 
 
  #ifdef _OPENMP
  #pragma omp parallel for private(i) firstprivate(IAccu) default(shared)
  #endif
  for(i = 1; i <= n; i++) 
  { 
    IAccu = 0.0; 
    accumulate(IAccu, T[i], ip);   
    zz[i] = rnd(IAccu); 
  } 
   
  end = GetTime();
  //cout << "Computation of [z]: " << end-start << endl;
 
// Resize 'xx' to length 'n'.
// ---------------------------------- 
   Resize(xx, n);  
  

// verified enclosure [x] for the absolute error of x. 
//-------------------------------------------------------------------------
   
//  VerificationStep(xx, zz, C, IsVerified);   // Attempt to compute [x] 
  start = GetTime();
  VerificationStep(xx, zz, C, IsVerified); 
  if (!IsVerified) 
      Err = VerivFailed; 
  else {
      //RefinementStep(xx, zz, C, Epsilon);
      xx = x + xx;             // The solution set lies in x+[x]
  }

   end = GetTime();
   //cout << "Verification: " << end-start << endl; 

  // Restore lower index bounds of 'Ap', 'bp' and ip 
  //-------------------------------------------------   
  SetLb(bp, ROW, bRowLow); SetLb(bp, COL, bColLow); 
  SetLb(Ap, ROW, ARowLow); SetLb(Ap, COL, AColLow); 
  SetLb(ip, ipLow); 
 
  opdotprec = oldprec;

} //ParLinSolve 


//---------------------------------------------------------------------------- 
// Purpose: The function 'ParLinSolve()' computes a verified enclosure for 
// the parametric solution set of a square parametric linear system of  
// equations A(p)*x=b(p), having a factorized by the parameters representation: 
//    (A0+[p1]A1+[p2]A2+ +[pk]Ak)x = (b0+[p1]b1+[p2]b2+ +[pk]bk).
// Parameters: 
//    In : 'Ap'         : coefficient matrix of the system in a factorized  
//                        representation Ap = (A0,A1,..,Ak) from R(n(k+1),n); 
//                        the matrix is passed as reference parameter to  
//                        save time for copying it; 
//         'bp'         : coefficients right-hand side of the system in a  
//                        factorized matrix representation 
//                        bp = (b0,b1,..bk) from R(n,k+1); matrix bp is passed  
//                        as reference parameter to save time for copying it; 
//         'ip'         : vector of the interval values for the parameters, 
//                        passed as reference parameter to save time for 
//                        copying it; 
//         'SharpC'     : flag signalling whether a sharp enclosure of the  
//                        contracting matrix C is to be computed; 
//         'Inner'      : flag signalling whether inner estimation of the hull
//			  of the solution set is to be computed;
//         'prec'       : precision to use for dot product computations
//                        (note: for performance reasons, some computations are
//                        always executed with precision 1 (double precision)
//    Out: 'xx'         : enclosure of the parametric solution set, resized to 
//                        standard index range with lower index bound 1; 
//	   'yy'		: inner estimation of the exact hull for the parametric
//			  solution set, resized to standard index range with 
//			  lower index bound 1;
//         'Err'        : error code; 
// Description: An approximate inverse 'R' of the mid-point matrix is computed  
//   by calling function 'MatInv()'. Then an approximate mid-point solution 'x' 
//   is computed applying a conventional real residual iteration. For the final 
//   verification, an interval residual iteration is performed. An enclosure of 
//   the unique solution is returned in the interval vector 'xx' and an inner  
//   approximation of the hull of the solution set is returned in the interval
//   vector 'yy'.
//---------------------------------------------------------------------------- 
 void ParLinSolveMain( rmatrix&  Ap, rmatrix&  bp, civector&  ip,  
                       civector&  xx, bool  SharpC, int&  Err, int prec ) 
{ 
  const int MaxResCorr = 10;    // Maximal number of real residual corrections 
  const real Epsilon   = 1e-10;
 
  int     p = VecLen(ip);            // Length of the parametric vector 'ip' 
				     // here it is of length p=p+1 for p parameters
  int     n = RowLen(Ap);            // Length of the rows of 'Ap' 
  int     m = ColLen(Ap);            // Length of the columns of 'Ap' 
  int     IsVerified, i, j, k; 
  int     ARowLow, AColLow,          // Lower index bounds of 'Ap', 'bp' 
          bRowLow, bColLow, ipLow;   // and ip 
  int     oldprec = opdotprec;
   
  cmatrix R(n,n);                         // To store the inverse of 'mid(A([ip]))' 
  double start,end;
   
  start = GetTime();

  if (m != n*p)                      // Error: 'A(p)' is not square 
    { Err = NotSquare; return; } 
 
  if (n != ColLen(bp) &&  
         p != RowLen(bp))            // Error: Dimensions of 'A(p)' and  
    { Err = DimensionErr; return; }  //       'b(p)' are not compatible 
 
 
// Start algorithm 
//---------------- 
  cvector       x(n), y(n), d(n);  // Allocate dynamic vectors 
  cimatrix       C(n, n);           // Allocate dynamic interval matrix C
  civector       zz(n);	           // Allocate dynamic nonparametric vector 
  cdotprecision  Accu;              // Allocate accumulators 
  cidotprecision IAccu; 
 
  Accu.set_dotprec(prec);
  IAccu.set_dotprec(prec);
  opdotprec = prec;

  C = 0.0;
  
// Normalize index range of 'Ap' and 'ip' to standard range 1..n 
// Save lower bounds of 'ip' and 'Ap' 
// to restore them before leaving 'ParLinSolve()'. 
//----------------------------------------------------------------- 
  ARowLow = Lb(Ap, ROW); SetLb(Ap, ROW, 1); 
  AColLow = Lb(Ap, COL); SetLb(Ap, COL, 1); 
  ipLow = Lb(ip); SetLb(ip, 1); 


 cmatrix    Am(n, n);               // Allocate dynamic mid-point matrix
 cvector    bm(n), pm(p);           // Allocate dynamic mid-point vectors
 
// Compute mid-point vectors/matrix: pm, Am, bm 
//---------------------------------------------------------------------- 
  pm = mid(ip); 
  bm = bp * pm;
  Am = Ap(1, n, 1, n); 
  for (i = 2; i <= p; i++)      
    Am += pm[i]* Ap(n*(i-1)+1, i*n, 1, n); 

//---------------------------------------------------------------------- 

  end = GetTime();
  //cout << "Preparation: " << end-start << endl;
 
  start = GetTime();
  MatInvAprx(Am, R, Err); 
  end = GetTime();
  //cout << "Computation of approximate inverse: " << end-start << endl;
  if (Err != NoError)                   // Error: Inversion failed 
    { Err = InvFailed; return; } 

// Normalize index range of 'R'  to standard index range 1..n 
//---------------------------------------------------------------------
  start = GetTime();
  SetLb(R, ROW, 1); SetLb(R, COL, 1);
 
  x = R * bm; k = 0;   // Real residual iteration 
  do  
  {                
    y = x; 
    for (i = 1; i <= n; i++)  
    {                               // Compute: d = #*(bm-Am*y) 
      Accu = bm[i];                 //----------------------- 
      accumulate(Accu, -Am[i], y); 
      d[i] = rnd(Accu); 
    } 
    
    for (i = 1; i <= n; i++)        // Compute: x = #*(y+R*d) 
    {                               //----------------------  
      Accu = y[i];                
      accumulate(Accu, R[i], d); 
      x[i] = rnd(Accu); 
    } 
     
    CheckForZeros(y, x); 
    k++; 
  }  
  while (!Accurate(y, x) && (k < MaxResCorr)); 
 
  end = GetTime();
  //cout << "Defect iteration: " << end-start << endl;

//  Resize mid-point vectors and matrix to save memory
//  ----------------------------------------------------------------
    Resize(pm);  Resize(bm);  Resize(Am);
    Resize(y);

// Prepare verification step: 

  start = GetTime();
  cimatrix T;	// Interval matrix for intermediate quantities
 
// compute enclosure [C]  of   I - R*A(p) for p in [ip]  
//--------------------------------------------------------------
 
  Accu.set_dotprec(1);
  IAccu.set_dotprec(1);
  opdotprec = 1;
  if (SharpC) {
      // Sharp enclosure C = I - R*A(p) 

      #ifdef CXSC_USE_BLAS

        for(i=1 ; i<=n ; i++)
           C[i][i] = 1.0;

        #ifdef _OPENMP
          #pragma omp parallel private(j,k) default(shared)
          {
            int size = n/omp_get_num_threads();

            #pragma omp for 
            for(j=1 ; j<=omp_get_num_threads() ; j++) {
              for(k=1 ; k<=p ; k++) {
                C((j-1)*size+1,j*size,1,n) -= ( R((j-1)*size+1, j*size, 1, n) * cimatrix(Ap(n*(k-1)+1, n*k, 1, n)) ) * ip[k];
              } 
            }
	
            //Compute rest if matrix not evenly distributable among threads
            #pragma omp single
            {
              if(n % omp_get_num_threads() != 0) {
                for(k=1 ; k<=p ; k++) {
                  C(size*omp_get_num_threads()+1,n,1,n) -= ( R(size*omp_get_num_threads()+1, n, 1, n) * cimatrix(Ap(n*(k-1)+1, n*k, 1, n)) ) * ip[k];
                }
              }
            }

          }

        #else 

          for(k=1 ; k<=p ; k++) 
             C -= (R * cimatrix(Ap(n*(k-1)+1, n*k, 1, n))) * ip[k];

        #endif //ifdef _OPENMP

 
      #else

        Resize(T, 1, p);       // Allocate dynamic temporary matrix T
  
        #ifdef _OPENMP
        #pragma omp parallel for private(i,j,k) firstprivate(Accu,IAccu,T) default(shared)
        #endif    
        for (i = 1; i <= n; i++) 
         for (j = 1; j <= n; j++) {                                     
           Accu = (i == j) ? 1.0 : 0.0;         
           accumulate(Accu, -R[i], Ap(1, n, 1, n)[Col(j)]); 
           T[1][1] = cinterval(rnd(Accu,RND_DOWN),rnd(Accu,RND_UP)); 
           for (k = 2;k <= p; k++) { 
             Accu = 0.0; 
	     accumulate(Accu, -R[i], Ap(n*(k-1)+1, n*k, 1, n)[Col(j)]); 
	     T[1][k] = cinterval(rnd(Accu,RND_DOWN),rnd(Accu,RND_UP)); 
           } 
           IAccu = 0.0; 
           accumulate(IAccu, T[Row(1)], ip); 
           C[i][j] = rnd(IAccu); 
         } 

      #endif

  } else {

      // Rough enclosure C = I - R*A([ip])  

      #ifdef CXSC_USE_BLAS

        Resize(T, n, n);     // Allocate dynamic temporary matrix T 
        T = Ap(1, n, 1, n);  //                  to contain A([ip])
  
        #ifdef _OPENMP
    
        #pragma omp parallel for private(i) default(shared)
        for(int r=1 ; r<=n ; r++) {
          for(int c=1 ; c<=n ; c++)
            for (i=2; i<=p; i++) 
                T[r][c] += ip[i]* Ap[r+n*(i-1)][c]; 
        }

        #else 

          for (i = 2; i <= p; i++)
          T += ip[i]* Ap(n*(i-1)+1, i*n, 1, n); 

        #endif //ifdef _OPENMP

        for(i=1 ; i<=n ; i++)
          C[i][i] = 1.0;

        C -= R*T;      

      #else

        Resize(T, n, n);     // Allocate dynamic temporary matrix T 
        T = Ap(1, n, 1, n);  //                  to contain A([ip])
        for (i = 2; i <= p; i++)
           T += ip[i]* Ap(n*(i-1)+1, i*n, 1, n); 

        #ifdef _OPENMP
        #pragma omp parallel for private(i,j) firstprivate(IAccu) default(shared)
        #endif
        for (i = 1; i <= n; i++) 
         for (j = 1; j <= n; j++) {                                       
           IAccu = (i == j) ? 1.0 : 0.0;         
           accumulate(IAccu, -R[i], T[Col(j)]); 
           rnd(IAccu, C[i][j]); 
         } 

      #endif //ifdef CXSC_USE_BLAS

  }
  Accu.set_dotprec(prec);
  IAccu.set_dotprec(prec);
  opdotprec = prec;

   end = GetTime();
   //cout << "Computation of [C]: " << end-start << endl;

// Resize T to length (n,p).
// Normalize index range of 'bp' to standard range 1..n.
// Save lower bounds of 'bp'
// to restore them before leaving 'ParLinSolve()'.
//---------------------------------------------------------------------
   Resize(T, n, p);
   bRowLow = Lb(bp, ROW); SetLb(bp, ROW, 1);
   bColLow = Lb(bp, COL); SetLb(bp, COL, 1);   
 
// compute enclosure [z] of z = R*(b(p)-A(p)*x). 
//---------------------------------------------------------------------
  start = GetTime();

  #ifdef _OPENMP
  #pragma omp parallel for private(k,i,j,m) firstprivate(Accu,IAccu,d,zz) default(shared)
  #endif
  for (k = 1; k <= p; k++) { 
 
   j = n*(k-1); 

    for (i = 1; i <= n; i++)               // Compute: d = #*(b_k - A_k * x) 
    {                                      //------------------------------ 
      Accu = bp[Col(k)][i]; 
      m = j + i;                 
      accumulate(Accu, -Ap[m], x);  
      d[i] = rnd(Accu); 
    } 
 
    for (i = 1; i <= n; i++)                // Compute: [y] = ##(b_k - A_k*x - d) 
    {                                       //----------------------------------  
      Accu = bp[Col(k)][i];      
      m = j + i;                             
      accumulate(Accu, -Ap[m], x); 
      Accu -= d[i]; 
      SetRe(zz[i], rnd(Re(Accu))); 
      SetIm(zz[i], rnd(Im(Accu))); 
    } 
 
                                //  Now b_k -A_k*x lies in d + [y].  
                                //  Thus R*(b_k-A_k*x) lies in [z]= ##(R*d + R*[y])     
    for (i = 1; i <= n; i++)  
    {   
      IAccu = 0.0; 
      accumulate(IAccu, R[i], d); 
      accumulate(IAccu, R[i], zz); 
      T[Col(k)][i] = rnd(IAccu); 
    } 
 
  }  // end for_k 
 
  #ifdef _OPENMP
  #pragma omp parallel for private(i) firstprivate(IAccu) default(shared)
  #endif
  for(i = 1; i <= n; i++) 
  { 
    IAccu = 0.0; 
    accumulate(IAccu, T[i], ip);   
    zz[i] = rnd(IAccu); 
  } 
   
  end = GetTime();
  //cout << "Computation of [z]: " << end-start << endl;
 
// Resize 'xx' to length 'n'.
// ---------------------------------- 
   Resize(xx, n);  
  

// verified enclosure [x] for the absolute error of x. 
//-------------------------------------------------------------------------
   
//  VerificationStep(xx, zz, C, IsVerified);   // Attempt to compute [x] 
  start = GetTime();
  VerificationStep(xx, zz, C, IsVerified); 
  if (!IsVerified) 
      Err = VerivFailed; 
  else {
      //RefinementStep(xx, zz, C, Epsilon);
      xx = x + xx;             // The solution set lies in x+[x]
  }

   end = GetTime();
   //cout << "Verification: " << end-start << endl; 

  // Restore lower index bounds of 'Ap', 'bp' and ip 
  //-------------------------------------------------   
  SetLb(bp, ROW, bRowLow); SetLb(bp, COL, bColLow); 
  SetLb(Ap, ROW, ARowLow); SetLb(Ap, COL, AColLow); 
  SetLb(ip, ipLow); 
 
  opdotprec = oldprec;

} //ParLinSolve 


//Entry function for solver
void  ParLinSolve (rmatrix& A, rmatrix& bp, ivector& ip, ivector& xx, bool SharpC, ivector& yy, int& err, int prec) {
  ParLinSolveMain(A,bp,ip,xx,SharpC,true,yy,err,prec);
}

//Entry function for solver
void  ParLinSolve (rmatrix& A, rmatrix& bp, ivector& ip, ivector& xx, bool SharpC, int& err, int prec) {
  ivector yy;
  ParLinSolveMain(A,bp,ip,xx,SharpC,false,yy,err,prec);
}

//Entry function for solver
void  ParLinSolve (cmatrix& A, cmatrix& bp, civector& ip, civector& xx, bool SharpC, int& err, int prec) {
  ParLinSolveMain(A,bp,ip,xx,SharpC,err,prec);
}

//Entry function for solver
void  ParLinSolve (rmatrix& A, rmatrix& bp, civector& ip, civector& xx, bool SharpC, int& err, int prec) {
  ParLinSolveMain(A,bp,ip,xx,SharpC,err,prec);
}
 
